"""This hold's the default config object used in the projet.

Overwrite it and pass it to the moving parts that depend on it.
The Config is a 'programmer config' intended for projects that want to implement
alpha zero for some game. It allows to do most stuff at runtime to allow for own policies
on resource use and own runtime config passed by the user."""
import time
import os


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# NN Server Settings
SELFPLAY_NN_SERVER_PORT = 5100
SELFEVAL_NN_SERVER_PORT = 5101
TRAINING_NN_SERVER_PORT = 5102

# Distribution settings
TRAINING_MASTER_PORT = 5200


class Configuration:
    def __init__(self):
        self._training_master_port = TRAINING_MASTER_PORT

        # Default Settings for Search/Training
        # Number of game states for one training batch
        self._training_batch_size = 64
        # Number of last games used for training
        self._training_history_size = 128
        # Simulations per selfplay/selfeval turn
        self._simulations_per_turn = 128
        # Turn time for each player during ai evaluation
        self._external_evaluation_turn_time = 1.0

        # Number of selfplay games for each iteration
        self._n_self_play = 42
        # Number of self evaluation games for each iteration, 0 will skip it
        self._n_self_eval = 0
        # Number of evaluation games against the ai-trivial client for each client, 0 will skip it
        self._n_external_eval = 0

        # The self evaluation avg. score needed to see this iteration as new best.
        # This means if the new weights scored >= this value they will be chosen as best weights.
        # 0.05 is a sane default for scores between -1 and 1.
        self._needed_avg_self_eval_score = 0.05

        # C-PUCT used in Tree search
        self._c_puct = 3

        # Configure the number of concurrent search threads, 1 means no multithreading
        self._n_search_threads_external_eval = 8
        self._n_search_threads_self_eval = 8
        self._n_search_threads_selfplay = 8

    def zmq_use_secure_connection(self):
        return False

    def zmq_client_secret(self):
        raise NotImplementedError('Return the path to the client secret!')

    def zmq_server_secret(self):
        raise NotImplementedError('Return the path to the server secret!')

    def zmq_server_public(self):
        raise NotImplementedError('Return the path to the server public!')

    def zmq_public_keys_dir(self):
        raise NotImplementedError('Return the path to the public key dir!')

    def nn_server_tensorboard_logdir(self, port, neural_network):
        return os.path.join(os.path.curdir, 'nn_logs/{}-{}'.format(port, round(time.time() * 1000)))

    def training_master_port(self):
        return self._training_master_port

    def nn_server_training_port(self):
        return TRAINING_NN_SERVER_PORT

    def nn_server_selfplay_port(self):
        return SELFPLAY_NN_SERVER_PORT

    def nn_server_selfeval_port(self):
        return SELFEVAL_NN_SERVER_PORT

    def training_batch_size(self):
        return self._training_batch_size

    def training_history_size(self):
        return self._training_history_size

    def simulations_per_turn(self):
        return self._simulations_per_turn

    def external_evaluation_turn_time(self):
        return self._external_evaluation_turn_time

    def n_self_play(self):
        return self._n_self_play

    def n_self_eval(self):
        return self._n_self_eval

    def n_external_eval(self):
        return self._n_external_eval

    def needed_avg_self_eval_score(self):
        return self._needed_avg_self_eval_score

    def external_ai_agent(self, start_game_state):
        """Returns an instance of an Agent class that plays for the external AI to compare to."""
        raise NotImplementedError()

    def external_evaluation_possible(self):
        """Specifies if the given machine can play games against the external ai client.
        This can come in handy if the external client only runs on specific machines, for example
        the client could only be available on Linux, but one machine runs Windows."""
        return True  # Usually this should work, maybe configure this at runtime

    def c_puct(self):
        return self._c_puct

    def n_search_threads_external_eval(self):
        return self._n_search_threads_external_eval

    def n_search_threads_self_eval(self):
        return self._n_search_threads_self_eval

    def n_search_threads_selfplay(self):
        return self._n_search_threads_selfplay



import hometrainer.core as core
from hometrainer.config import Configuration
import time
import concurrent.futures
import numpy as np


class Agent:
    """Agent that is able to play a game. This means it can take an game state and give it's best move.

    The agent get's some more lifecycle callbacks if it needs to for example note moves of the enemy to
    improve its decisions. An agent is usually an AI program that calculates a good move,
    but can also be an interface for an human player to compete in an match.

    Note: Agents must be able to be copied, as they might be initialized multiple times for one match
          with more than two players. This helps you if your agent holds player specific state."""
    def game_start(self, game_state, agents_player):
        """Called when the new game starts. Can be used for setup code."""
        pass

    def find_move_with_time_limit(self, game_state, move_time):
        """Implement your logic to find a move with a given time limit in here."""
        raise NotImplementedError('Move finding (time limit) method not implemented!')

    def find_move_with_iteration_limit(self, game_state, move_iterations):
        """Implement your logic to find a move with a given iteration limit in here."""
        raise NotImplementedError('Move finding (iteration limit) method not implemented!')

    def move_executed(self, old_game_state, move, new_game_state):
        """Called after a move was executed in the current game."""
        pass

    def game_ended(self, game_state):
        """Called when the game ended. Can be used for teardown code."""
        pass


class NeuralNetworkAgent(Agent):
    """Agent using a neural network and the alpha zero search for finding moves."""
    def __init__(self, nn_client, config: Configuration, collect_evaluations=False, temperature=None):
        """
        :param nn_client: The neural network client used to get evaluations
        :param collect_evaluations: If true a evaluations will be created using the search trees from each executed move
        :param temperature: If set to a value moves will be chosen probabilistic using this temperature
        """
        self.nn_client = nn_client

        self.collect_evaluations = collect_evaluations
        self.collected_evaluations = []

        self.temperature = temperature
        self.config = config

        self.current_mcts_node = None
        self.thread_pool = None

    def game_start(self, game_state, agents_player):
        self.current_mcts_node = core.MCTSNode(1.0, game_state, self.config)

        n_threads = self.config.n_search_threads_selfplay()
        self.thread_pool = None
        if n_threads > 1:
            self.thread_pool = concurrent.futures.ThreadPoolExecutor(n_threads)

        self.nn_client.start(self.config)

    def find_move_with_time_limit(self, game_state, move_time):
        move_end_time = time.time() + move_time

        while True:
            self._run_mcts_simulations(16)
            if move_end_time < time.time():
                break

        self._create_evaluation()
        return self._find_best_move()

    def find_move_with_iteration_limit(self, game_state, move_iterations):
        self._run_mcts_simulations(move_iterations)
        self._create_evaluation()
        return self._find_best_move()

    def _run_mcts_simulations(self, n_simulations):
        # We can run serial or parallel in a thread pool
        if not self.thread_pool:
            for i in range(n_simulations):
                self.current_mcts_node.run_simulation_step(self.nn_client)
        else:
            futures = []
            for i in range(n_simulations):
                futures.append(self.thread_pool.submit(self.current_mcts_node.run_simulation_step, self.nn_client))
            concurrent.futures.wait(futures)

    def _find_best_move(self):
        # Either select probabilistic or simply take the best move.
        if self.temperature:
            move_probabilities = self.current_mcts_node.move_probabilities(self.temperature).items()
            moves = [item[0] for item in move_probabilities]
            probabilities = [item[1] for item in move_probabilities]

            # Select the move according to the probability distribution
            index = np.random.choice(len(moves), p=probabilities)
            return moves[index]
        else:
            move_probabilities = self.current_mcts_node.move_probabilities(1.0)
            best_move = None
            best_prob = -1

            for move, prob in move_probabilities.items():
                if prob > best_prob:
                    best_prob = prob
                    best_move = move

            return best_move

    def _create_evaluation(self):
        """Creates an evaluation of the current MCTSExecutor and adds it to the collected evaluations for this run"""
        if not self.collect_evaluations:
            return

        evaluation = self.current_mcts_node.game_state.wrap_in_evaluation()
        evaluation.set_move_probabilities(self.current_mcts_node.move_probabilities(1.0))

        self.collected_evaluations.append(evaluation)

    def move_executed(self, _, move, new_game_state):
        # Try to keep parts of the tree if possible
        if self.current_mcts_node and self.current_mcts_node.children and self.current_mcts_node.children[move]:
            self.current_mcts_node = self.current_mcts_node.children[move]
        else:
            self.current_mcts_node = core.MCTSNode(1.0, new_game_state, self.config)

    def game_ended(self, game_state):
        # If we did collect evaluations update them according to the actual results.
        actual_game_result = game_state.calculate_scores()
        for evaluation in self.collected_evaluations:
            evaluation.set_expected_result(actual_game_result)

        # Tear down any resources left
        self.thread_pool.shutdown(wait=False)

        # Stop NN client
        self.nn_client.stop()



import pickle
import copy as old_copy
import zmq.auth
from hometrainer.config import Configuration
from zmq.auth.thread import ThreadAuthenticator
import tempfile
import os
import shutil
import zipfile
import io
import numpy as np


def copy(obj):
    return old_copy.copy(obj)


def deepcopy(obj):
    """Way faster then normal deep copy."""
    return pickle.loads(pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL))


def secure_client_connection(client, ctx, config: Configuration, only_localhost=False):
    """Configures certificates for zeromq client connection."""
    if not config.zmq_use_secure_connection():
        return

    _init_auth(ctx, only_localhost, config)

    client_public, client_secret = zmq.auth.load_certificate(config.zmq_client_secret())
    client.curve_secretkey = client_secret
    client.curve_publickey = client_public

    server_public, _ = zmq.auth.load_certificate(config.zmq_server_public())
    # The client must know the server's public key to make a CURVE connection.
    client.curve_serverkey = server_public


def secure_server_connection(server, ctx, config: Configuration, only_localhost=False):
    """Configures certificates for zeromq server connection."""
    if not config.zmq_use_secure_connection():
        return

    _init_auth(ctx, only_localhost, config)

    server_public, server_secret = zmq.auth.load_certificate(config.zmq_server_secret())
    server.curve_secretkey = server_secret
    server.curve_publickey = server_public
    server.curve_server = True  # must come before bind


def save_neural_net_to_zip_binary(neural_network, session):
    with tempfile.TemporaryDirectory() as base_dir:
        # Prepare Saving Path's
        checkpoint_dir = os.path.join(base_dir, 'checkpoint')
        checkpoint_file = os.path.join(checkpoint_dir, 'checkpoint.ckpt')
        os.makedirs(checkpoint_dir)

        # Save it
        neural_network.save_weights(session, checkpoint_file)

        # Pack it up to send it through the network
        checkpoint_zip = os.path.join(base_dir, 'checkpoint')
        shutil.make_archive(checkpoint_zip, 'zip', checkpoint_dir)
        with open(checkpoint_zip + '.zip', 'rb') as file:
            return file.read()


def load_neural_net_from_zip_binary(zip_binary, neural_network, session):
    with tempfile.TemporaryDirectory() as base_dir:
        # Prepare Saving Path's
        checkpoint_dir = os.path.join(base_dir, 'checkpoint')
        checkpoint_file = os.path.join(checkpoint_dir, 'checkpoint.ckpt')
        checkpoint_zip = os.path.join(base_dir, 'checkpoint.zip')
        os.makedirs(checkpoint_dir)

        # Prepare the checkpoint data for loading
        with open(checkpoint_zip, 'wb') as file:
            file.write(zip_binary)

        with zipfile.ZipFile(checkpoint_zip, "r") as zip_ref:
            zip_ref.extractall(checkpoint_dir)

        neural_network.load_weights(session, checkpoint_file)


def _init_auth(ctx, only_localhost, config: Configuration):
    auth = ThreadAuthenticator(ctx)
    auth.start()

    if only_localhost:
        auth.allow('127.0.0.1')
    else:
        auth.allow('*')

    auth.configure_curve(domain='*', location=config.zmq_public_keys_dir())


def count_files(dir_path):
    return len([name for name in os.listdir(dir_path) if os.path.isfile(os.path.join(dir_path, name))])


def plot_external_eval_avg_score(work_dir, lower_bound, upper_bound, show_progress_lines,
                                 smoothing=0.1, min=-1, max=1, middle=0,
                                 x_label='Batch', y_label='Average Game Outcome',
                                 smoothed_score_lable='Smoothed Score', score_label='Score',
                                 x_scaling=1, new_was_better_label='New weights better than previous'):
    """Returns a binary containing a plot of the current winrate."""
    import matplotlib.pyplot as plt
    import hometrainer.distribution as distribution
    import matplotlib.patches as mpatches

    x_scaling = x_scaling

    progress = distribution.TrainingRunProgress(os.path.join(work_dir, 'stats.json'))
    progress.load_stats()

    avg_score = []
    x_steps = []

    new_was_better = []
    for iteration in progress.stats.iterations[lower_bound:upper_bound]:
        if iteration.ai_eval.n_games < 14:
            break

        avg_score.append(iteration.ai_eval.nn_score / iteration.ai_eval.n_games)
        x_steps.append(iteration.ai_eval.end_batch * x_scaling)

        if iteration.self_eval.new_better:
            new_was_better.append(iteration.ai_eval.end_batch * x_scaling)

    smoothed_wins = smooth_array(avg_score, smoothing)

    label_1, = plt.plot(x_steps, avg_score, linestyle='--', label=score_label)
    label_2, = plt.plot(x_steps, smoothed_wins, label=smoothed_score_lable)
    plt.ylim(min, max)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.axhline(y=middle, color='r')
    if show_progress_lines:
        label_3 = None
        for better_x in new_was_better:
            label_3 = plt.axvline(x=better_x, linestyle=':', label=new_was_better_label)
        if label_3:
            plt.legend(handles=[label_1, label_2, label_3])
        else:
            plt.legend(handles=[label_1, label_2])
    else:
        plt.legend(handles=[label_1, label_2])

    result = io.BytesIO()
    plt.savefig(result, dpi=400)
    plt.clf()

    return result


def smooth_array(array, smoothing):
    smoothed_array = []
    max_steps = max(1, round(len(array) * smoothing))
    for i in range(len(array)):
        values = [array[i]]
        for j in range(1, max_steps + 1):
            if i - j < 0:
                break
            values.append(array[i - j])
        for j in range(1, max_steps + 1):
            if i + j >= len(array):
                break
            values.append(array[i + j])

        smoothed_array.append(np.average(values))

    return smoothed_array




import hometrainer.agents
from hometrainer.config import Configuration
import threading
import numpy as np
import hometrainer.util as util
import os
import pickle
import random
import functools
import time
import hometrainer.core as core
import logging
import hometrainer.agents as agents


class GameExecutor:
    """Handles the simulation of one game between different opponents.

    This includes mapping these opponents to players in the game,
    progressing the game according to the opponents move decisions
    and evaluating the game outcome.

    Effectively this can be used as:
    Play a game with this starting game state between these participants and tell me the average score of each agent."""
    def __init__(self, start_game_state, agents, add_up_time=False):
        self.start_game_state = start_game_state
        self.external_agents = agents

        # If set to true players can save up time from past moves.
        self.add_up_time = add_up_time

    def play_game(self, time_limit=None, iteration_limit=None):
        """Plays one game and returns the average scores of each agent."""
        if not time_limit and not iteration_limit:
            raise ValueError("At least one of 'time_limit' or 'iteration_limit' must be provided!")

        self._shuffle_player_mapping()

        time_left = dict()
        for player in self.start_game_state.get_player_list():
            time_left[player] = 0
            agent = self.player_executor_mappings[player]
            agent.game_start(self.start_game_state, player)

        current_game_state = self.start_game_state
        while not current_game_state.is_finished():
            next_player = current_game_state.get_next_player()
            current_agent = self.player_executor_mappings[next_player]

            if time_limit:
                time_left[next_player] = time_left[next_player] + time_limit
                if self.add_up_time:
                    final_time_limit = time_left[next_player]
                else:
                    final_time_limit = time_limit

                before_move = time.time()
                move = current_agent.find_move_with_time_limit(current_game_state, final_time_limit)
                after_move = time.time()

                time_left[next_player] = time_left[next_player] - (after_move - before_move)
            else:
                move = current_agent.find_move_with_iteration_limit(current_game_state, iteration_limit)

            next_game_state = current_game_state.execute_move(move)
            for agent in self.internal_agents:
                agent.move_executed(current_game_state, move, next_game_state)

            current_game_state = next_game_state

        for agent in self.internal_agents:
            agent.game_ended(current_game_state)

        player_scores = current_game_state.calculate_scores()
        agent_scores = {agent: 0 for agent in self.external_agents}
        for player, score in player_scores.items():
            current_agent = self.player_result_mappings[player]
            agent_scores[current_agent] = agent_scores[current_agent] + score

        avg_agent_scores = dict()
        for agent, score in agent_scores.items():
            avg_agent_scores[agent] = score / self.agent_counts[agent]

        return avg_agent_scores

    def _shuffle_player_mapping(self):
        players = self.start_game_state.get_player_list()

        # Shuffle the agents to give no one an advantage
        np.random.shuffle(self.external_agents)
        self.agent_counts = {agent: 0 for agent in self.external_agents}
        self.player_result_mappings = dict()
        self.player_executor_mappings = dict()

        # Assign agents to each player in this game
        for i in range(len(players)):
            agent = self.external_agents[i % len(self.external_agents)]
            self.player_result_mappings[players[i]] = agent
            self.player_executor_mappings[players[i]] = util.deepcopy(agent)
            self.agent_counts[agent] = self.agent_counts[agent] + 1

        # Keep track of them as our 'internal' players
        self.internal_agents = [agent for player, agent in self.player_executor_mappings.items()]


class SelfplayExecutor:
    """Handles the simulation one selfplay game.

    This should run one game of selfplay and return a list of all states and all
    corresponding probability/value targets that can then be used as training data."""
    def __init__(self, game_state, nn_clients, n_simulations_per_move, config: Configuration, temperature=1.0):
        self.config = config
        self.start_game_state = game_state
        self.n_simulations_per_move = n_simulations_per_move
        self.nn_clients = nn_clients
        self.temperature = temperature

    def run(self):
        """Actually run the selfplay game.

        This will run the one game played against itself."""
        agents = [hometrainer.agents.NeuralNetworkAgent(nn_client, self.config, True, self.temperature)
                  for nn_client in self.nn_clients]
        game_executor = GameExecutor(self.start_game_state, agents)
        game_executor.play_game(iteration_limit=self.n_simulations_per_move)

        # Return all collected evaluation by each neural network agent.
        # This allows to have multiple agents play against each other and collect different stats, helping
        # with searching a diverse set of game states.
        return [evaluation for agent in game_executor.internal_agents for evaluation in agent.collected_evaluations]


class TrainingExecutor:
    """Manages the training process of a neural network.

    This is managing the training data and the training process.
    The class is given neural network client to work with.

    The training history size indicates how many of the last games to consider
    for training (e.g. use the 500 most recent games of training data)."""
    def __init__(self, nn_client, data_dir, training_history_size, apply_transformations=True):
        super().__init__()
        self.nn_client = nn_client
        self.training_history_size = training_history_size

        self.apply_transformations = apply_transformations

        # We will keep the training and test data in a local folder.
        # This class is only responsible for somehow doing the training,
        # this does not constrain it to run only on this machine,
        # but its a good start to have all training data somewhere for whatever training method.
        self.data_dir = data_dir
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

        self.lock = threading.Lock()
        self._current_number = util.count_files(self.data_dir)

    def add_examples(self, evaluations):
        with self.lock:
            self._current_number = self._current_number + 1

            with open(os.path.join(self.data_dir, "{0:010d}.pickle".format(self._current_number)), 'wb') as file:
                pickle.dump(evaluations, file)

    def get_examples(self, n_examples):
        with self.lock:
            evaluations = []
            while len(evaluations) < n_examples:
                oldest_index = max(1, self._current_number - self.training_history_size)
                number = random.randint(oldest_index, self._current_number)

                try:
                    loaded_evaluations = self._load_evaluations(number)
                except IOError:
                    continue

                random.shuffle(loaded_evaluations)
                end_index = min(round(n_examples / 8 + 1), len(loaded_evaluations))
                evaluations = evaluations + loaded_evaluations[:end_index]

            return evaluations

    # TODO: Make cache size adjustable
    @functools.lru_cache(maxsize=512)
    def _load_evaluations(self, example_number):
        with open(os.path.join(self.data_dir, "{0:010d}.pickle".format(example_number)), 'rb') as file:
            loaded_evaluations = pickle.load(file)

            # Add every possible transformation to our samples
            if self.apply_transformations:
                results = []
                for evaluation in loaded_evaluations:
                    for i in evaluation.get_possible_transformations():
                        transformed_evaluation = evaluation.apply_transformation(i)
                        results.append(transformed_evaluation)

                return results
            else:
                return loaded_evaluations

    def load_weights(self, filename):
        with open(filename, 'rb') as file:
            weights_zip_binary = file.read()
            self.nn_client.load_weights(weights_zip_binary)

    def save_weights(self, filename):
        weights_zip_binary = self.nn_client.save_weights()
        with open(filename, 'wb') as file:
            file.write(weights_zip_binary)

    def run_training_batch(self, batch_size=32):
        # Skip if there is no data
        if self._current_number <= 0:
            time.sleep(10)
            return

        evaluations = self.get_examples(batch_size)
        self.nn_client.execute_training_batch(evaluations)

    def log_loss(self, epoch, batch_size=32):
        raise NotImplementedError()


class ModelEvaluator:
    """Compares two neural network configurations by playing out a game between them.

    This is a optional step in the training process where we play with our current best network weights
    against our currently trained weights to see if they are better and therefore our new best weights.
    This step in the training is optional."""
    def __init__(self, nn_client_one, nn_client_two, start_game_state: core.GameState, config: Configuration):
        self.nn_client_one = nn_client_one
        self.nn_client_two = nn_client_two
        self.start_game_state = start_game_state
        self.config = config

    def run(self, n_simulations):
        """Executes the match between the two neural networks.

        Returns an array with [avg_score_nn_one, avg_score_nn_two]."""
        try:
            agent_one = hometrainer.agents.NeuralNetworkAgent(self.nn_client_one, self.config)
            agent_two = hometrainer.agents.NeuralNetworkAgent(self.nn_client_two, self.config)
            agents = [agent_one, agent_two]

            game_executor = GameExecutor(self.start_game_state, agents)
            agent_scores = game_executor.play_game(iteration_limit=n_simulations)

            return [agent_scores[agent_one], agent_scores[agent_two]]
        except Exception:
            import traceback
            traceback.print_exc()


class ExternalEvaluator:
    """Compares a neural network to an external AI program by playing out a small tournament."""
    def __init__(self, nn_client, external_agent, start_game_state, config: Configuration):
        self.nn_agent = agents.NeuralNetworkAgent(nn_client, config)
        self.external_agent = external_agent
        self.start_game_state = start_game_state
        self.config = config

    def run(self, turn_time):
        try:
            return self._run_internal(turn_time)
        except Exception as e:
            logging.error("Exception during external evaluation match! {}".format(e))
            import traceback
            traceback.print_exc()
            # We do not care too much if there was an error.
            # This is just a metric used to see 'general' progress, one game wont matter too much.
            # We are so careful here, as this is a place with potential code that accesses external network
            # resources, so stuff can go really wrong (for example the ReversiXT client can not resume at the current
            # point in the match after a crash
            return [0, 0]

    def _run_internal(self, turn_time):
        game_executor = GameExecutor(self.start_game_state, [self.nn_agent, self.external_agent])
        agent_scores = game_executor.play_game(time_limit=turn_time)

        return [agent_scores[self.nn_agent], agent_scores[self.external_agent]]



import pickle
import multiprocessing
import numpy as np
import logging
import importlib
import zmq
import time
import random
import threading
import hometrainer.core as core
import hometrainer.util as util
from hometrainer.config import Configuration
import tensorflow as tf



class NeuralNetwork:
    """Wrapper that represents a single neural network instance.

    This is intended to abstract away the actual creation, training and execution of the neural network.
    This should hopefully also allow to re-use major parts of the code for different network structures.

    The network is not responsible for managing its scope/tensorflow graph, this is managed in the
    nn_server module and abstracts away the actual tf session and the interaction with it.

    See examples for how to implement these, but mostly all functions follow their brief descriptions."""
    def construct_network(self, sess, graph):
        raise NotImplementedError("Add the construction of your custom graph structure.")

    def init_network(self):
        raise NotImplementedError("Run initialisation code for your network.")

    def input_conversion_function(self):
        raise NotImplementedError("Return a reference to the function converting evaluations to input for thin NN.")

    def output_conversion_function(self):
        raise NotImplementedError("Return a reference to the function filling outputs into evaluations.")

    def execute_batch(self, sess, input_arrays):
        raise NotImplementedError("Add implementation that takes prepared input arrays and executes them as a batch.")

    def train_batch(self, sess, input_arrays, output_arrays):
        raise NotImplementedError("Add implementation that executes one batch training step.")

    def save_weights(self, sess, filename):
        raise NotImplementedError("Add implementation that saves the weights of this network to a checkpoint.")

    def load_weights(self, sess, filename):
        raise NotImplementedError("Add implementation that loads the weights of this network to a checkpoint.")

    def log_training_progress(self, sess, tf_file_writer, input_arrays, target_arrays, training_batch):
        raise NotImplementedError("Add implementation to write stats on the current training progress.")


class NeuralNetworkClient:
    """Client Functionality for our neural network abstraction.

    The neural network is abstracted into an client and server, in the way that the actual network can run as
    a 'server' and that the client acts as a simple proxy/interface to use the neural network.
    This helps, as we can then simply call functions on the neural network in our alpha zero code,
    without worrying how it is actually executed, it's a plain function call.

    The biggest drawback is, that we have to serialize/deserialize requests to the network.
    This might be a problem, but can be solved by using foster deserialization like ray.
    One extra bonus of decoupling the neural network is that we have no synchronisation issues and
    that the network runs in it's own process, so it can use extra cpu cores without dragging down the
    main program."""
    def __init__(self, address):
        self.address = address

        self.context = None
        self.nn_server_socket = None
        self.internal_router = None

        self.input_conversion = None
        self.output_conversion = None

        # Make sure we will never get collisions. The name is arbitrary anyways,
        # as it is only used to handle concurrent use of the neural network client.
        self.internal_address = 'inproc://nn_client_router_{}_{}'.format(time.time(), random.random)

    def evaluate_game_state(self, game_state: core.GameState) -> core.Evaluation:
        """Executes a neural network to evaluate a given game state.

        Returns the evaluation generated by the network.
        Might block for a while because other evaluations are being performed."""
        # Create the evaluation to send to the NN.
        # We will apply rotations on random instances.
        evaluation = game_state.wrap_in_evaluation()
        evaluation = evaluation.convert_to_normal()
        evaluation = evaluation.apply_transformation(random.randint(0, 6))

        # Execute it using the neural network process.
        # We hand it through our internal proxy to be able to
        # map parallel execution results properly.
        input_array, target_array = self.input_conversion(evaluation)
        response = self.send_to_server(ExecutionRequest(input_array))
        evaluation = self.output_conversion(evaluation, response.output_array)

        # Undo our random rotation on the instance.
        evaluation = evaluation.undo_transformations()
        evaluation = evaluation.convert_from_normal()

        return evaluation

    def execute_training_batch(self, evaluations):
        """Executes a batch of training using the given evaluations."""
        inputs = []
        targets = []
        for evaluation in evaluations:
            input_array, target_array = self.input_conversion(evaluation, calculate_target=True)
            inputs.append(input_array)
            targets.append(target_array)

        return self.send_to_server(TrainingRequest(np.asarray(inputs), np.asarray(targets)))

    def save_weights(self):
        """Saves the current weights to a binary of it's checkpoint."""
        response = self.send_to_server(SaveWeightsRequest())
        return response.checkpoint_data

    def load_weights(self, checkpoint_zip_binary):
        """Loads a checkpoint from the binary of it's checkpoint."""
        self.send_to_server(LoadWeightsRequest(checkpoint_zip_binary))

    def shutdown_server(self):
        return self.send_to_server(ShutdownRequest())

    def send_to_server(self, message):
        """Sends a message to the NN Server and returns its response message.

        This method is thread save and will block each thread until the individual answer
        has been returned by the server (e.g. for execution requests this can take until
        a full batch was executed)"""
        client = self.context.socket(zmq.REQ)
        client.connect(self.internal_address)

        client.send_pyobj(message)
        response = client.recv_pyobj()

        client.close()

        return response

    def start(self, config: Configuration):
        self.context = zmq.Context()

        self.nn_server_socket = self.context.socket(zmq.DEALER)
        util.secure_client_connection(self.nn_server_socket, self.context, config)
        self.nn_server_socket.connect(self.address)

        self.internal_router = self.context.socket(zmq.ROUTER)
        self.internal_router.bind(self.internal_address)

        def run_proxy():
            try:
                zmq.proxy(self.internal_router, self.nn_server_socket)
            except zmq.ZMQError:
                pass  # The closed proxy is on purpose, so ignore the error.

        threading.Thread(target=run_proxy).start()

        # Get the conversion methods
        response = self.send_to_server(ConversionFunctionRequest())
        self.input_conversion = response.input_conversions
        self.output_conversion = response.output_conversions

    def stop(self):
        self.internal_router.setsockopt(zmq.LINGER, 0)
        self.internal_router.close()
        self.nn_server_socket.setsockopt(zmq.LINGER, 0)
        self.nn_server_socket.close()
        self.context.term()


def start_nn_server(port, nn_class_name, config: Configuration, nn_init_args=(), batch_size=32, log_dir=None, start_batch=0):
    """Starts a neural network server on the given port.

    The server is started using the venv cpython executable.
    The nn_class_name must be the fully qualified name of the neural network class.
    This method can be used from any python interpreter, for example also from
    the pypy interpreter, as all tensorflow components are loaded in the new process."""
    logging.info('Starting NN Server on port {}.'.format(port))
    p = multiprocessing.Process(target=_start_nn_server_internal,
                                args=(port, nn_class_name, batch_size, log_dir, start_batch, nn_init_args, config))
    p.start()
    return p


def _start_nn_server_internal(port, nn_class_name, batch_size, log_dir, start_batch,
                              nn_init_args, config: Configuration):
    module_name, class_name = nn_class_name.rsplit('.', 1)
    nn_module = importlib.import_module(module_name)
    nn_class = getattr(nn_module, class_name)

    server = NeuralNetworkServer(port, nn_class(*nn_init_args), batch_size,
                                 config=config, start_batch=start_batch, log_dir=log_dir)
    server.run()


# Message types used to communicate with the NNServer
class AbstractMessage:
    pass


class Response(AbstractMessage):
    def __init__(self, response_ids):
        self.response_ids = response_ids

    def to_multipart(self):
        return self.response_ids + [b'', pickle.dumps(self)]


class ShutdownRequest(AbstractMessage):
    pass


class SaveWeightsRequest(AbstractMessage):
    pass


class SaveWeightsResponse(Response):
    def __init__(self, response_ids, checkpoint_data):
        super().__init__(response_ids)
        self.checkpoint_data = checkpoint_data


class LoadWeightsRequest(AbstractMessage):
    def __init__(self, checkpoint_data):
        self.checkpoint_data = checkpoint_data


class ConversionFunctionRequest(AbstractMessage):
    pass


class ConversionFunctionResponse(Response):
    def __init__(self, response_ids, input_conversion, output_conversion):
        super().__init__(response_ids)
        self.input_conversions = input_conversion
        self.output_conversions = output_conversion


class ExecutionRequest(AbstractMessage):
    def __init__(self, input_array):
        self.input_array = input_array


class ExecutionResponse(Response):
    def __init__(self, response_ids, input_array):
        super().__init__(response_ids)
        self.input_array = input_array
        self.output_array = None

    def set_output(self, output_array):
        self.output_array = output_array
        self.input_array = None


class TrainingRequest(AbstractMessage):
    def __init__(self, input_arrays, target_arrays):
        self.input_arrays = input_arrays
        self.target_arrays = target_arrays


class NeuralNetworkServer:
    """The server component of our decoupled neural network.

    This actually runs the neural network in a separate process.
    See the nn_client for details."""
    def __init__(self, port, neural_network, batch_size, log_dir: None, config: Configuration, start_batch=0):
        self.port = port
        self.neural_network = neural_network
        self.input_conversion = neural_network.input_conversion_function()
        self.output_conversion = neural_network.output_conversion_function()
        self.stopped = False
        self.socket = None
        self.execution_responses = []
        self.batch_size = batch_size
        self.config = config

        if log_dir:
            self.log_dir = log_dir
        else:
            self.log_dir = config.nn_server_tensorboard_logdir(port, neural_network)

        self.n_batches_for_log = 100
        self.current_training_batch = start_batch
        self.log_file_writer = None
        self.graph = None

    def run(self):
        # Init network code. Router is used because we want to synchronize
        # all request in in a request-reply fashion.
        context = zmq.Context()
        self.socket = context.socket(zmq.ROUTER)
        util.secure_server_connection(self.socket, context, self.config)
        self.socket.bind('tcp://*:{}'.format(self.port))

        # Shutdown gracefully in case of interrupts
        try:
            # Setup a tensorflow session to be used for the whole run.
            self.graph = tf.Graph()
            with self.graph.as_default():
                # GPU Memory is allocated only as needed, this allows to run multiple
                # sessions on one machine.
                config = tf.ConfigProto()
                config.gpu_options.allow_growth = True

                with tf.Session(config=config) as sess:
                    self.neural_network.construct_network(sess, self.graph)
                    self.neural_network.init_network()

                    while not self.stopped:
                        try:
                            # Multipart messages are needed to correctly map the response
                            message = self.socket.recv_multipart(flags=zmq.NOBLOCK)
                            response_ids = message[0:-2]
                            message_content = pickle.loads(message[-1])

                            # Process the incoming message...
                            if isinstance(message_content, ShutdownRequest):
                                self._process_shutdown_request(response_ids, message_content)
                            elif isinstance(message_content, ExecutionRequest):
                                self._process_execution_request(response_ids, message_content, sess)
                            elif isinstance(message_content, TrainingRequest):
                                self._process_training_request(response_ids, message_content, sess)
                            elif isinstance(message_content, SaveWeightsRequest):
                                self._process_save_weights_request(response_ids, message_content, sess)
                            elif isinstance(message_content, LoadWeightsRequest):
                                self._process_load_weights_request(response_ids, message_content, sess)
                            elif isinstance(message_content, ConversionFunctionRequest):
                                self._process_conversion_function_request(response_ids, message_content)
                            else:
                                print("Unknown message '{}' received!".format(message_content))

                        except zmq.ZMQError:
                            # Also execute not full batches if no new data arrived in time
                            if len(self.execution_responses) >= 1:
                                self._execute_batch(sess)
                            else:
                                # Don't  busy wait all the time
                                time.sleep(0.01)
                if self.log_file_writer:
                    self.log_file_writer.close()
        except KeyboardInterrupt:
            logging.error('Keyboard Interrupt, shutting down NN server...')
        finally:
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close()
            context.term()

    def _process_execution_request(self, response_ids, message_content, sess):
        response = ExecutionResponse(response_ids, message_content.input_array)
        self.execution_responses.append(response)

        if len(self.execution_responses) > self.batch_size:
            self._execute_batch(sess)

    def _execute_batch(self, sess):
        inputs = [execution_response.input_array for execution_response in self.execution_responses]
        outputs = self.neural_network.execute_batch(sess, inputs)

        for i in range(len(outputs)):
            self.execution_responses[i].set_output(outputs[i])
            self.socket.send_multipart(self.execution_responses[i].to_multipart())

        self.execution_responses = []

    def _process_training_request(self, response_ids, message_content, sess):
        if not self.log_file_writer:
            self.log_file_writer = tf.summary.FileWriter(self.log_dir, self.graph)

        if self.current_training_batch % self.n_batches_for_log == 0:
            # It's not smart to test with the training data.
            # But as we do not really have a test set here and we
            # will only train on this exact data after the summary data it seems 'ok' to do for now.
            self.neural_network.log_training_progress(sess, self.log_file_writer, message_content.input_arrays,
                                                      message_content.target_arrays, self.current_training_batch)

        # FIXME: Research what causes this call to leak memory if the inputs get too big.
        #        The leak happens at a fixed threshold on the input size (about 200).
        self.neural_network.train_batch(sess, message_content.input_arrays, message_content.target_arrays)
        self.current_training_batch += 1

        response = Response(response_ids)
        self.socket.send_multipart(response.to_multipart())

    def _process_save_weights_request(self, response_ids, message_content, sess):
        checkpoint_content = util.save_neural_net_to_zip_binary(self.neural_network, sess)
        response = SaveWeightsResponse(response_ids, checkpoint_content)
        self.socket.send_multipart(response.to_multipart())

    def _process_load_weights_request(self, response_ids, message_content, sess):
        util.load_neural_net_from_zip_binary(message_content.checkpoint_data, self.neural_network, sess)

        response = Response(response_ids)
        self.socket.send_multipart(response.to_multipart())

    def _process_conversion_function_request(self, response_ids, message_content):
        response = ConversionFunctionResponse(response_ids, self.input_conversion, self.output_conversion)
        self.socket.send_multipart(response.to_multipart())

    def _process_shutdown_request(self, response_ids, message_content):
        self.stopped = True

        response = Response(response_ids)
        self.socket.send_multipart(response.to_multipart())



from setuptools import setup

setup(name='hometrainer',
      version='0.1',
      description='Simple AlhpaZero implementation',
      url='https://github.com/FritzFlorian/hometrainer',
      author='Fritz Florian',
      license='MIT',
      packages=['hometrainer'],
      install_requires=[
          'tensorflow',
          'pyzmq',
          'numpy',
          'py-kim',
          'matplotlib'
      ],
      zip_safe=False)



"""Code that coordinates different steps defined in the core module.

This is coordinating the training process and managing the results.
The distribution module allows one training master and many playing slaves to generate data.

This is the highest level interface to use the """
from hometrainer.config import Configuration
import multiprocessing
import zmq
import zmq.error
import pickle
import logging
import time
import numpy as np
import enum
import os
import threading
import kim
import json
import signal
import traceback
import hometrainer.util as util
import hometrainer.executors as executors
import hometrainer.neural_network as neural_network


class PlayingSlave:
    """Runs selfplay games and evaluations using a given neural network configuration.

    The selfplay server reports it's results back to a master node.
    The master node coordinates workloads, stores results and configures the worker."""
    TIMEOUT = 30

    class WorkResult:
        pass

    class EmptyWorkResult(WorkResult):
        pass

    class SelfplayWorkResult(WorkResult):
        def __init__(self, evaluation_lists):
            self.evaluation_lists = evaluation_lists

    class SelfEvaluationWorkResult(WorkResult):
        def __init__(self, nn_one_score, nn_two_score, n_games):
            self.nn_one_score = nn_one_score
            self.nn_two_score = nn_two_score
            self.n_games = n_games

    class AIEvaluationWorkResult(WorkResult):
        def __init__(self, nn_score, ai_score, n_games):
            self.nn_score = nn_score
            self.ai_score = ai_score
            self.n_games = n_games

    class WorkRequest:
        def __init__(self, work_result):
            self.work_result = work_result

    class SelfplayWorkResponse:
        def __init__(self, n_games, nn_name, weights_zip_binary, game_states, simulations_per_turn):
            self.n_games = n_games
            self.nn_name = nn_name
            self.weights_zip_binary = weights_zip_binary
            self.game_states = game_states
            self.simulations_per_turn = simulations_per_turn

    class SelfEvaluationWorkResponse:
        def __init__(self, n_games, nn_name, weights_binary_one, weights_binary_two,
                     game_states, simulations_per_turn, epoch):
            self.n_games = n_games
            self.nn_name = nn_name
            self.weights_binary_one = weights_binary_one
            self.weights_binary_two = weights_binary_two
            self.game_states = game_states
            self.simulations_per_turn = simulations_per_turn
            self.epoch = epoch

    class AIEvaluationWorkResponse:
        def __init__(self, n_games, nn_name, weights_zip_binary, game_states, turn_time, epoch):
            self.n_games = n_games
            self.nn_name = nn_name
            self.weights_zip_binary = weights_zip_binary
            self.game_states = game_states
            self.turn_time = turn_time
            self.epoch = epoch

    class WaitResponse:
        def __init__(self, wait_time):
            self.wait_time = wait_time

    def __init__(self, master_address, config: Configuration=None):
        if not config:
            # Use default configuration...
            config = Configuration()
        self.config = config

        self.master_address = master_address

        self.nn_client_one = None
        self.nn_client_host_one = ''
        self.nn_class_name_one = ''

        self.nn_client_two = None
        self.nn_client_host_two = ''
        self.nn_class_name_two = ''

        self.context = None
        self.zmq_client = None
        self.poll = None

        # Disable Interrupts in workers, we shutdown gracefully by our self
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.process_pool = multiprocessing.Pool(multiprocessing.cpu_count() - 1)
        signal.signal(signal.SIGINT, original_sigint_handler)

    def run(self):
        self.context = zmq.Context()
        self.poll = zmq.Poller()

        try:
            self._handle_connections()
        except KeyboardInterrupt:
            logging.error('Keyboard Interrupt, shutting down Playing Slave...')
        finally:
            self.process_pool.terminate()
            if self.nn_client_one:
                self.nn_client_one.stop()
            if self.nn_client_two:
                self.nn_client_two.stop()

            self._disconnect_client()

    def _handle_connections(self):
        last_work_result = self.EmptyWorkResult()

        while True:
            self._connect_client()

            try:
                logging.info('Sending work request to master server...')
                self.zmq_client.send_pyobj(self.WorkRequest(last_work_result), flags=zmq.NOBLOCK)

                socks = dict(self.poll.poll(round(self.TIMEOUT * 1000)))
                if socks.get(self.zmq_client) == zmq.POLLIN:
                    response = self.zmq_client.recv()
                    if response:
                        last_work_result = self._handle_response(response)
                    else:
                        raise zmq.ZMQError()
                else:
                    raise zmq.ZMQError()
            except zmq.ZMQError:
                last_work_result = self.EmptyWorkResult()
                self._disconnect_client()
                logging.info('Server connection closed. Waiting {} seconds, then reconnect...'.format(self.TIMEOUT))
                time.sleep(self.TIMEOUT)

    def _handle_response(self, response):
        message = pickle.loads(response)
        if isinstance(message, self.WaitResponse):
            time.sleep(message.wait_time)
            return self.EmptyWorkResult()
        if isinstance(message, self.SelfplayWorkResponse):
            logging.info('Start processing Selfplay work request from master...')
            self._prepare_networks(message.nn_name, message.weights_zip_binary)
            evaluation_lists = self._play_games(message.n_games, message.game_states, message.simulations_per_turn)
            return self.SelfplayWorkResult(evaluation_lists)
        if isinstance(message, self.SelfEvaluationWorkResponse):
            logging.info('Start processing SelfEvaluation work request from master...')
            self._prepare_networks(message.nn_name, message.weights_binary_one, message.weights_binary_two)
            scores = self._self_evaluate(message.n_games, message.game_states, message.simulations_per_turn)
            return self.SelfEvaluationWorkResult(scores[0], scores[1], message.n_games)
        if isinstance(message, self.AIEvaluationWorkResponse):
            logging.info('Start processing AIEvaluation work request from master...')
            if self.config.external_evaluation_possible():
                self._prepare_networks(message.nn_name, message.weights_zip_binary)
                scores = self._ai_evaluate(message.n_games, message.game_states, message.turn_time)
                return self.AIEvaluationWorkResult(scores[0], scores[1], message.n_games)
            else:
                logging.warning('Can not process work request, AITrivial executable is missing!')
                logging.warning('Waiting 30 seconds and retry work request...')
                time.sleep(30)
                return self.EmptyWorkResult()

        return self.EmptyWorkResult()

    def _prepare_networks(self, nn_class_name, weights_zip_binary, weights_zip_binary_two=None):
        if self.nn_class_name_one != nn_class_name:
            self._restart_network_one(nn_class_name)
        if self.nn_class_name_two != nn_class_name:
            self._restart_network_two(nn_class_name)

        if weights_zip_binary:
            self.nn_client_one.load_weights(weights_zip_binary)
        if weights_zip_binary_two:
            self.nn_client_two.load_weights(weights_zip_binary_two)

    def _restart_network_one(self, nn_class_name):
        self.nn_class_name_one = nn_class_name
        if self.nn_client_one:
            self.nn_client_one.shutdown_server()
            self.nn_client_one.stop()
            time.sleep(15)

        neural_network.start_nn_server(self.config.nn_server_selfplay_port(), self.nn_class_name_one, self.config)
        self.nn_client_host_one = 'tcp://localhost:{}'.format(self.config.nn_server_selfplay_port())
        self.nn_client_one = neural_network.NeuralNetworkClient(self.nn_client_host_one)
        self.nn_client_one.start(self.config)

    def _restart_network_two(self, nn_class_name):
        self.nn_class_name_two = nn_class_name
        if self.nn_client_two:
            self.nn_client_two.shutdown_server()
            self.nn_client_two.stop()
            time.sleep(15)

        neural_network.start_nn_server(self.config.nn_server_selfeval_port(), self.nn_class_name_two, self.config)
        self.nn_client_host_two = 'tcp://localhost:{}'.format(self.config.nn_server_selfeval_port())
        self.nn_client_two = neural_network.NeuralNetworkClient(self.nn_client_host_two)
        self.nn_client_two.start(self.config)

    def _play_games(self, n_games, game_states, simulations_per_turn):
        results = []
        for _ in range(n_games):
            nn_executor_client = neural_network.NeuralNetworkClient(self.nn_client_host_one)
            game_state = np.random.choice(game_states)

            params = (game_state, nn_executor_client, simulations_per_turn, self.config)
            result = self.process_pool.apply_async(PlayingSlave._play_game, params,
                                                   callback=PlayingSlave.selfplay_callback)
            results.append(result)

        evaluation_lists = []
        for result in results:
            evaluation_lists.append(result.get())

        return evaluation_lists

    @staticmethod
    def _play_game(game_state, nn_executor_client, n_simulations, config):
        try:
            selfplay_executor = executors.SelfplayExecutor(game_state, [nn_executor_client], n_simulations, config)
            result = selfplay_executor.run()
        except Exception as e:
            # Print and re-raise the exception, as python will ignore it otherwise
            print(traceback.format_exc())
            raise e

        return result

    @staticmethod
    def selfplay_callback(result):
        logging.info('Selfplay-Game finished.')

    def _self_evaluate(self, n_games, game_states, simulations_per_turn):
        results = []
        for _ in range(n_games):
            nn_executor_client_one = neural_network.NeuralNetworkClient(self.nn_client_host_one)
            nn_executor_client_two = neural_network.NeuralNetworkClient(self.nn_client_host_two)

            game_state = np.random.choice(game_states)

            params = (game_state, nn_executor_client_one, nn_executor_client_two, simulations_per_turn, self.config)
            result = self.process_pool.apply_async(PlayingSlave._play_evaluation_game, params,
                                                   callback=PlayingSlave.eval_callback)
            results.append(result)

        total_score = [0, 0]
        for result in results:
            score = result.get()
            total_score[0] += score[0]
            total_score[1] += score[1]

        return total_score

    @staticmethod
    def _play_evaluation_game(game_state, nn_executor_one, nn_executor_two, n_simulations, config):
        model_evaluator = executors.ModelEvaluator(nn_executor_one, nn_executor_two, game_state, config)
        result = model_evaluator.run(n_simulations)

        return result

    @staticmethod
    def eval_callback(result):
        logging.info('Evaluation-Game finished. ({}:{})'.format(result[0], result[1]))

    def _ai_evaluate(self, n_games, game_states, turn_time):
        results = []
        for _ in range(n_games):
            nn_executor_client_one = neural_network.NeuralNetworkClient(self.nn_client_host_one)

            game_state = np.random.choice(game_states)

            params = (game_state, nn_executor_client_one, turn_time, self.config)
            result = self.process_pool.apply_async(PlayingSlave._play_ai_evaluation_game, params,
                                                   callback=PlayingSlave.ai_eval_callback)
            results.append(result)

        total_score = [0, 0]
        for result in results:
            score = result.get()
            for i in range(2):
                total_score[i] += score[i]

        return total_score

    @staticmethod
    def _play_ai_evaluation_game(game_state, nn_executor_one, turn_time, config: Configuration):
        ai_evaluator = executors.ExternalEvaluator(nn_executor_one, config.external_ai_agent(game_state),
                                                   game_state, config)
        result = ai_evaluator.run(turn_time)

        return result

    @staticmethod
    def ai_eval_callback(result):
        score = result
        logging.info('AI-Evaluation-Game finished. ({}:{})'.format(score[0], score[1]))

    def _connect_client(self):
        if not self.zmq_client:
            self.zmq_client = self.context.socket(zmq.REQ)
            util.secure_client_connection(self.zmq_client, self.context, self.config)
            self.zmq_client.connect(self.master_address)

            self.poll.register(self.zmq_client, zmq.POLLIN)

    def _disconnect_client(self):
        if self.zmq_client:
            self.zmq_client.setsockopt(zmq.LINGER, 0)
            self.zmq_client.close()
            self.poll.unregister(self.zmq_client)
            self.zmq_client = None


class TrainingMaster:
    """Master Node that coordinates the training process.

    This will order slaves to execute selfplay, selfevaluation and aievaluation games,
    collect the results and use them to train a neural network instance."""
    class State(enum.Enum):
        SELFPLAY = 'SELFPLAY'
        SELFEVAL = 'SELFEVAL'
        AIEVAL = 'AIEVAL'

    DATA_DIR = 'selfplay-data'
    WEIGHTS_DIR = 'weights-history'
    BEST_WEIGHTS = 'best-weights.zip'
    LOG_DIR = 'tensorboard-logs'

    def __init__(self, work_dir, nn_name, start_game_states, config: Configuration=None):
        if not config:
            # Use default configuration...
            config = Configuration()
        self.config = config

        self.work_dir = work_dir
        if not os.path.exists(self.work_dir):
            os.makedirs(self.work_dir)
        self.weights_dir = os.path.join(self.work_dir, self.WEIGHTS_DIR)
        if not os.path.exists(self.weights_dir):
            os.makedirs(self.weights_dir)
        self.log_dir = os.path.join(self.work_dir, self.LOG_DIR, 'current-run')
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.nn_name = nn_name
        self.port = config.training_master_port()

        self.start_game_states = start_game_states

        self.server = None
        self.context = None

        self.training_executor = None
        self.nn_client = None

        # weights cache
        self.current_weights_binary = None
        self.best_weights_binary = None

        # training threads
        self.stopped = False
        self.training_thread_one = None
        self.training_thread_two = None
        self.training_progress_lock = threading.Lock()

        # keep some stats about our progress
        self.progress = TrainingRunProgress(os.path.join(work_dir, 'stats.json'))
        self.progress.load_stats()

        # We could customize settings here, but defaults are fine
        self._copy_stats(config)

        self.progress.save_stats()

    def _copy_stats(self, config: Configuration):
        self.progress.stats.settings.turn_time = config.external_evaluation_turn_time()
        self.progress.stats.settings.needed_avg_score = config.needed_avg_self_eval_score()
        self.progress.stats.settings.n_ai_eval = config.n_external_eval()
        self.progress.stats.settings.simulations_per_turn = config.simulations_per_turn()
        self.progress.stats.settings.n_self_eval = config.n_self_eval()
        self.progress.stats.settings.n_self_play = config.n_self_play()
        self.progress.stats.settings.batch_size = config.training_batch_size()
        self.progress.stats.settings.training_history_size = config.training_history_size()

    def run(self):
        self.context = zmq.Context()
        self.server = self.context.socket(zmq.REP)
        util.secure_server_connection(self.server, self.context, self.config)
        self.server.bind('tcp://*:{}'.format(self.port))

        self._setup_nn()
        self._setup_training_executor()
        self._setup_training_threads()
        try:
            self._handle_messages()
        except KeyboardInterrupt:
            logging.error('Keyboard Interrupt, shutting down Training Master...')
        finally:
            self.progress.save_stats()

            self.stopped = True
            self.nn_client.stop()
            self.server.setsockopt(zmq.LINGER, 0)
            self.server.close()

    def _setup_nn(self):
        neural_network.start_nn_server(self.config.nn_server_training_port(), self.nn_name, self.config,
                                  log_dir=self.log_dir, start_batch=self.progress.stats.progress.current_batch)
        self.nn_client = neural_network.NeuralNetworkClient('tcp://localhost:{}'.format(self.config.nn_server_training_port()))
        self.nn_client.start(self.config)

        if self.progress.stats.progress.iteration > 0:
            checkpoint_name = 'checkpoint-{0:05d}.zip'.format(self.progress.stats.progress.iteration)
            with open(os.path.join(self.weights_dir, checkpoint_name), 'rb') as file:
                weights_binary = file.read()
                self.nn_client.load_weights(weights_binary)

            with open(os.path.join(self.work_dir, 'best-checkpoint.zip'), 'rb') as file:
                self.best_weights_binary = file.read()
        else:
            self.progress.stats.progress.iteration = 1
            self.best_weights_binary = self.nn_client.save_weights()
            self._save_best_weights()

        self.current_weights_binary = self.nn_client.save_weights()
        self._save_current_weights()
        self.progress.save_stats()

    def _setup_training_executor(self):
        self.training_executor = executors.TrainingExecutor(self.nn_client, os.path.join(self.work_dir, self.DATA_DIR),
                                                            self.progress.stats.settings.training_history_size)

    def _setup_training_threads(self):
        self.training_thread_one = threading.Thread(target=self._run_training_loop)
        self.training_thread_one.start()
        self.training_thread_two = threading.Thread(target=self._run_training_loop)
        self.training_thread_two.start()

    def _run_training_loop(self):
        while not self.stopped:
            try:
                self.training_executor.run_training_batch(self.progress.stats.settings.batch_size)
            except zmq.error.ContextTerminated:
                return
            self._add_training_progress()

    def _add_training_progress(self):
        with self.training_progress_lock:
            self.progress.stats.progress.current_batch += 1
            if self.progress.stats.progress.current_batch % 100 == 0:
                self.progress.save_stats()

    def _handle_messages(self):
        while True:
            request = self.server.recv_pyobj()

            if isinstance(request, PlayingSlave.WorkRequest):
                logging.debug('Work request {}.'.format(request))
                self._handle_work_request(request)
            else:
                self.server.send('unsupported message type')

    def _handle_work_request(self, request: PlayingSlave.WorkRequest):
        work_result = request.work_result

        if isinstance(work_result, PlayingSlave.SelfplayWorkResult):
            self._handle_selfplay_result(work_result)
        if isinstance(work_result, PlayingSlave.SelfEvaluationWorkResult):
            self._handle_selfeval_result(work_result)
        if isinstance(work_result, PlayingSlave.AIEvaluationWorkResult):
            self._handle_aieval_result(work_result)

        self._send_work_response()

    def _send_work_response(self):
        n_games = 7
        simulations_per_turn = self.progress.stats.settings.simulations_per_turn
        epoch = self.progress.stats.progress.iteration
        turn_time = self.progress.stats.settings.turn_time
        game_states = self.start_game_states

        if self.progress.stats.progress.state == self.State.SELFPLAY:
            logging.info('Sending Selfplay Work Request to Playing Slave')
            self.server.send_pyobj(
                PlayingSlave.SelfplayWorkResponse(n_games, self.nn_name, self.best_weights_binary, game_states,
                                                  simulations_per_turn))
        if self.progress.stats.progress.state == self.State.SELFEVAL:
            logging.info('Sending Selfeval Work Request to Playing Slave')
            self.server.send_pyobj(
                PlayingSlave.SelfEvaluationWorkResponse(n_games, self.nn_name, self.current_weights_binary,
                                                        self.best_weights_binary, game_states, simulations_per_turn,
                                                        epoch))
        if self.progress.stats.progress.state == self.State.AIEVAL:
            logging.info('Sending AI Eval Work Request to Playing Slave')
            self.server.send_pyobj(
                PlayingSlave.AIEvaluationWorkResponse(n_games, self.nn_name, self.best_weights_binary, game_states,
                                                      turn_time, epoch))

    def _handle_selfplay_result(self, work_result):
        logging.info('Getting Selfplay Results from Playing Slave')
        n_evaluations = len(work_result.evaluation_lists)

        self.progress.stats.current_epoch().self_play.n_games += n_evaluations
        if self.progress.stats.progress.add_samples(self.State.SELFPLAY, n_evaluations):
            self._progress_to_selfeval()

            # Save a snapshot of the current weights
            self.current_weights_binary = self.nn_client.save_weights()
            self._save_current_weights()

        for evaluations in work_result.evaluation_lists:
            self.training_executor.add_examples(evaluations)

    def _progress_to_selfeval(self):
        # Progress to the next step of the training
        self.progress.stats.progress.n_remaining = self.progress.stats.settings.n_self_eval
        self.progress.stats.progress.state = self.State.SELFEVAL

        # Collect some stats on the current training step
        self.progress.stats.current_epoch().self_play.end_batch = self.progress.stats.progress.current_batch
        self.progress.stats.current_epoch().self_eval.start_batch = self.progress.stats.progress.current_batch

        if self.config.n_self_eval() <= 0:
            # Skip SelvEval
            logging.info('Skipping SelfEval...')

            # Be sure to still select the new weights as best ones
            self._select_new_weights_as_best()

            self._progress_to_ai_eval()
        else:
            logging.info('Start Self-Evaluation for Iteration {}...'.format(self.progress.stats.progress.iteration))

    def _progress_to_ai_eval(self):
        # Progress to the next step of the training
        self.progress.stats.progress.n_remaining = self.progress.stats.settings.n_ai_eval
        self.progress.stats.progress.state = self.State.AIEVAL

        # Collect some stats on the current training step
        self.progress.stats.current_epoch().self_eval.end_batch = self.progress.stats.progress.current_batch
        self.progress.stats.current_epoch().ai_eval.start_batch = self.progress.stats.progress.current_batch

        if self.config.n_external_eval() <= 0:
            # Skip external ai eval
            logging.info('Skipping External AI Eval...')
            self._progress_to_selfplay()
        else:
            logging.info('Start AI-Evaluation for Iteration {}...'.format(self.progress.stats.progress.iteration))

    def _progress_to_selfplay(self):
        # Progress to the next step of the training
        self.progress.stats.progress.n_remaining = self.progress.stats.settings.n_self_play
        self.progress.stats.progress.state = self.State.SELFPLAY
        self.progress.stats.progress.iteration += 1

        # Collect some stats on the current training step
        self.progress.stats.current_epoch().ai_eval.end_batch = self.progress.stats.progress.current_batch
        self.progress.stats.current_epoch().self_play.start_batch = self.progress.stats.progress.current_batch

    def _save_current_weights(self):
        checkpoint_name = 'checkpoint-{0:05d}.zip'.format(self.progress.stats.progress.iteration)
        with open(os.path.join(self.weights_dir, checkpoint_name), 'wb') as file:
            file.write(self.current_weights_binary)

    def _save_best_weights(self):
        with open(os.path.join(self.work_dir, 'best-checkpoint.zip'), 'wb') as file:
            file.write(self.best_weights_binary)

    def _handle_selfeval_result(self, work_result):
        logging.info('Getting Selfeval Results from Playing Slave (New {} vs. Old {})'.format(work_result.nn_one_score, work_result.nn_two_score))

        self.progress.stats.current_epoch().self_eval.n_games += work_result.n_games
        self.progress.stats.current_epoch().self_eval.old_score += work_result.nn_two_score
        self.progress.stats.current_epoch().self_eval.new_score += work_result.nn_one_score

        if self.progress.stats.progress.add_samples(self.State.SELFEVAL, work_result.n_games):
            self._progress_to_ai_eval()

            # See if the new nn was better
            old_score = self.progress.stats.current_epoch().self_eval.old_score
            new_score = self.progress.stats.current_epoch().self_eval.new_score
            n_games = self.progress.stats.current_epoch().self_eval.n_games

            logging.info('Finishing selfplay with result {} vs. {}'
                         .format(new_score, old_score))
            if new_score / n_games > self.progress.stats.settings.needed_avg_score:
                logging.info('Choosing new weights, as it scored better then the current best.')
                self.progress.stats.current_epoch().self_eval.new_better = True
                self.best_weights_binary = self.current_weights_binary

                self._save_best_weights()
            else:
                logging.info('Choosing old weights, as the new ones where not better.')
                self.progress.stats.current_epoch().self_eval.new_better = False

    def _select_new_weights_as_best(self):
        self.progress.stats.current_epoch().self_eval.new_better = True
        self.best_weights_binary = self.current_weights_binary

        self._save_best_weights()

    def _handle_aieval_result(self, work_result):
        logging.info('Getting Selfeval Results from Playing Slave (NN {} vs. AI {})'.format(work_result.nn_score, work_result.ai_score))

        self.progress.stats.current_epoch().ai_eval.n_games += work_result.n_games
        self.progress.stats.current_epoch().ai_eval.ai_score += work_result.ai_score
        self.progress.stats.current_epoch().ai_eval.nn_score += work_result.nn_score

        if self.progress.stats.progress.add_samples(self.State.AIEVAL, work_result.n_games):
            self._progress_to_selfplay()

            logging.info('Start Selfplay for Iteration {}...'.format(self.progress.stats.progress.iteration))
            self._save_current_weights()


class TrainingRunProgress:
    """Captures all important information about one training run.

    This includes the settings for the run as well as it's progress.
    This can also be used to coordinate the run itself, including continuing a run."""
    def __init__(self, stats_file_name):
        self.stats_file_name = stats_file_name
        self.stats = TrainingRunStats()
        self.stats.progress.n_remaining = self.stats.settings.n_self_play

    def save_stats(self):
        # Store the enum as an string.
        # Could be cleaned up by using an appropriate mapper.
        to_save = util.deepcopy(self.stats)
        to_save.progress.state = self.stats.progress.state.value

        mapper = TrainingRunStatsMapper(obj=to_save)
        json_string = json.dumps(mapper.serialize(), indent=4)

        with open(self.stats_file_name + '-copy', 'w') as stats_file:
            stats_file.write(json_string)
        with open(self.stats_file_name, 'w') as stats_file:
            stats_file.write(json_string)

    def load_stats(self):
        if os.path.isfile(self.stats_file_name):
            with open(self.stats_file_name, 'r') as stats_file:
                json_data = json.loads(stats_file.read())
                mapper = TrainingRunStatsMapper(data=json_data)
                loaded = mapper.marshal()
                loaded.progress.state = TrainingMaster.State(loaded.progress.state)

                self.stats = loaded


class TrainingRunStats:
    """Data container for the actual training run stats."""
    class Settings:
        def __init__(self):
            """Init's with default settings. Overwrite if needed."""
            # Number of game states for one training batch
            self.batch_size = 64
            # Number of last games used for training
            self.training_history_size = 128
            # Simulations per selfplay/selfeval turn
            self.simulations_per_turn = 128
            # Turn time for each player during ai evaluation
            self.turn_time = 1.0

            # Number of selfplay games for each iteration
            self.n_self_play = 42
            # Number of self evaluation games for each iteration
            self.n_self_eval = 21
            # Number of evaluation games against the ai-trivial client for each client
            self.n_ai_eval = 14

            # The self evaluation avg. score needed to see this iteration as new best
            self.needed_avg_score = 0.05

    class Progress:
        def __init__(self):
            """Progress statistics of one training run."""
            # Our current state
            self.state = TrainingMaster.State.SELFPLAY
            # The number of samples of the current state needed to progress to the next state
            self.n_remaining = 0
            # The current iteration
            self.iteration = 0
            # The current batch
            self.current_batch = 0

        def add_samples(self, state, n_samples):
            """Call when new samples arrive. Returns True if progress to the next state should be made."""
            if self.state == state:
                self.n_remaining -= n_samples
                if self.n_remaining <= 0:
                    return True

            return False

    class Iteration:
        class SelfEval:
            pass
        class SelfPlay:
            pass
        class AIEval:
            pass

        def __init__(self):
            self.self_eval = self.SelfEval()
            self.self_eval.n_games = 0
            self.self_eval.old_score = 0
            self.self_eval.new_score = 0
            self.self_eval.start_batch = 0
            self.self_eval.end_batch = 0
            self.self_eval.new_better = False
            self.self_eval.avg_score = 0

            self.self_play = self.SelfPlay()
            self.self_play.start_batch = 0
            self.self_play.end_batch = 0
            self.self_play.n_games = 0

            self.ai_eval = self.AIEval()
            self.ai_eval.n_games = 0
            self.ai_eval.start_batch = 0
            self.ai_eval.end_batch = 0
            self.ai_eval.nn_score = 0
            self.ai_eval.ai_score = 0

    def __init__(self):
        self.settings = self.Settings()
        self.progress = self.Progress()
        self.iterations = []

    def current_epoch(self):
        while len(self.iterations) < self.progress.iteration:
            self.iterations.append(self.Iteration())

        return self.iterations[self.progress.iteration - 1]


# Python has NO good json serializer, so we need some boilerplate
class SettingsMapper(kim.Mapper):
    __type__ = TrainingRunStats.Settings
    batch_size = kim.field.Integer()
    training_history_size = kim.field.Integer()
    simulations_per_turn = kim.field.Integer()
    turn_time = kim.field.Float()
    n_self_play = kim.field.Integer()
    n_self_eval = kim.field.Integer()
    n_ai_eval = kim.field.Integer()
    needed_avg_score = kim.field.Float()


class ProgressMapper(kim.Mapper):
    __type__ = TrainingRunStats.Progress
    state = kim.field.String()
    n_remaining = kim.field.Integer()
    iteration = kim.field.Integer()
    current_batch = kim.field.Integer()


class IterationSelfEvalMapper(kim.Mapper):
    __type__ = TrainingRunStats.Iteration.SelfEval
    n_games = kim.field.Integer()
    old_score = kim.field.Integer()
    new_score = kim.field.Integer()
    start_batch = kim.field.Integer()
    end_batch = kim.field.Integer()
    new_better = kim.field.Boolean()

class IterationSelfPlayMapper(kim.Mapper):
    __type__ = TrainingRunStats.Iteration.SelfPlay
    n_games = kim.field.Integer()
    start_batch = kim.field.Integer()
    end_batch = kim.field.Integer()


class IterationAIEvalMapper(kim.Mapper):
    __type__ = TrainingRunStats.Iteration.AIEval
    n_games = kim.field.Integer()
    start_batch = kim.field.Integer()
    end_batch = kim.field.Integer()
    nn_score = kim.field.Integer()
    ai_score = kim.field.Integer()


class IterationMapper(kim.Mapper):
    __type__ = TrainingRunStats.Iteration
    self_eval = kim.field.Nested(IterationSelfEvalMapper, allow_create=True)
    ai_eval = kim.field.Nested(IterationAIEvalMapper, allow_create=True)
    self_play = kim.field.Nested(IterationSelfPlayMapper, allow_create=True)


class TrainingRunStatsMapper(kim.Mapper):
    __type__ = TrainingRunStats
    settings = kim.field.Nested(SettingsMapper, allow_create=True)
    progress = kim.field.Nested(ProgressMapper, allow_create=True)
    iterations = kim.field.Collection(kim.field.Nested(IterationMapper, allow_create=True), allow_create=True)



"""Holds all code needed for the actual alpha zero search and training.

This is the 'core' code that executes the actual parts needed/described in the alpha zero
papers. The actual search is executed in the MCTSNode, all other classes coordinate the process.

This core could be used by itself to execute the training, but usually it will be used by the
distribution module to handle the 'large scale' coordination of the process. Also this implementation
is fully generic, it needs an actual game implementation to work."""
import threading
import math
import hometrainer.util as util
from hometrainer.config import Configuration


class Move:
    """Represents an executed move.

    This is some sort of 'short' description
    on what the move was. In chess this would for example be 'white, Ke4'.
    This will be used to refer to individual moves in all context, e.g. when telling how good a move is."""
    def __hash__(self):
        raise NotImplementedError()

    def __eq__(self, other):
        raise NotImplementedError()

    def __ne__(self, other):
        return not (self == other)


class GameState:
    """Holds all information needed to describe one specific game state.

    This includes information on turn limits, active player and so on.
    Must also remember the last executed move (when using the execution logic)."""

    def get_next_game_states(self):
        """Returns an array of game states that follow this one after one move.

        The returned array is empty if the game is finished.
        The self object must not be changed by this call! (at least not publicly visible, caching is fine)"""
        raise NotImplementedError()

    def get_next_player(self):
        """Return the next player that will perform an move."""
        raise NotImplementedError()

    def is_finished(self):
        """Return if the game is finished. Overwrite with a faster method if needed."""
        return len(self.get_next_game_states()) == 0

    def get_last_move(self):
        """Returns the last executed move as a Move object."""
        raise NotImplementedError()

    def execute_move(self, move: Move):
        """Return the game state following from the given move."""
        raise NotImplementedError()

    def get_player_list(self):
        """Returns a list of all players.

        This is usually a list of enums or numbers. These will be used to associate statistics to players.
        The players used here must be the same as used in the Move class!

        Example for two player game: [1, 2] or ['player_one', 'player_two']"""
        raise NotImplementedError()

    def get_average_score(self):
        """The average outcome of a game.

        This is needed as a starting value for not processed nodes in the search tree.
        0 is a sane default scores are between -1 and 1."""
        return 0

    def get_virtual_loss(self):
        """The value used for virtual loss in the search tree.

        Keep this reasonable considering your scoring scheme.
        -1 is a sane default for scores between -1 and 1"""
        return -1

    def calculate_scores(self):
        """Score the game state according to the game rules.

        This will be called on terminal game states. Scores should be somehow normalized."""
        raise NotImplementedError()

    def wrap_in_evaluation(self):
        """Return an evaluation that wraps this game state.

        Simply call the constructor of your custom evaluation for this."""
        raise NotImplementedError()


class Evaluation:
    """The 'glue' between GameState and the neural Network.

    Holds move probabilities and the expected game outcome returned by the neural network.
    This will work for both directions: Training the neural network and using it to evaluate game states.
    Think of it as one training sample (input: BoardState, output: move probs & expected outcome).

    To work properly this will also encapsulate the logic needed to transform a board state, so that the
    current active play is always player one (to have a simpler neural network) and to apply rotations/mirrors
    to the game state if appropriate in the given game."""

    def __init__(self, game_state: GameState):
        self.game_state = game_state

    def get_expected_result(self):
        """Returns an floating point value on the expected result for all players in an dictionary.

        This is the guess of the neural network for the game outcome for this player.
        The output should be some sort of normalized score, higher is better,
        best values between 0 and 1 or -1 and 1 from worst to best player."""
        raise NotImplementedError()

    def set_expected_result(self, expected_results):
        """Same as getter, simple the matching setter"""
        raise NotImplementedError()

    def get_move_probabilities(self):
        """Returns a directory with the form {move: move_probability} of all possible moves in the game state."""
        raise NotImplementedError()

    def set_move_probabilities(self, move_probabilities):
        """Set a directory with the form {move: move_probability} of all possible moves in the game state."""
        raise NotImplementedError()

    def convert_to_normal(self):
        """Converts the evaluation to a form where the next active player is player one.

        Returns the converted evaluation and does not change the original data!
        DO NOT change self or self.game_state in the process. This is to avoid any side effects!
        If the evaluation is already in normal form it is simply returned."""
        raise NotImplementedError()

    def convert_from_normal(self):
        """Converts the evaluation from normal form to its original form.

        Returns the converted evaluation and does not change the original data!
        DO NOT change self or self.game_state in the process. This is to avoid any side effects!
        If the evaluation is already in normal form it is simply returned."""
        raise NotImplementedError()

    def get_possible_transformations(self):
        """List of all possible transformations (mirror, rotate, ...).

        This will be used to alter the game_state for more training samples."""
        # By default we do not support any transformations.
        # If possible implement this as soon as the 'basics' work, as it helps a lot.
        return ['identity']

    def apply_transformation(self, transformation):
        """Returns the transformed evaluation and does not change the original data!

        The transformed evaluation should keep track of all applied transformations, so it can undo them later on.
        This should really transform all internal data, including the move probabilities and expected game result"""
        # By default we do not support any transformations.
        # If possible implement this as soon as the 'basics' work, as it helps a lot.
        return self

    def undo_transformations(self):
        """Undos all previously applied transformations.

        Returns the converted evaluation and does not change the original data!"""
        # By default we do not support any transformations.
        # If possible implement this as soon as the 'basics' work, as it helps a lot.
        return self


class MCTSNode:
    """A single node in the constructed search tree.

    Each node holds its visit count, its value, its direct child nodes and the probability of the move
    that lead to them. This is 'the other way around' of what the paper states (values per edge), but was
    simpler to implement and follow, as most programmers are used to hold information in the actual nodes
    and model connections as references.
    It also simplifies where values belong, as the expected outcome of a game state is in the node holding it.

    This is the 'heart' of the actual alpha zero search. It is actually quite simple and easy to follow."""

    def __init__(self, probability, game_state: GameState, config: Configuration):
        """ Create a new node for the search tree.

        :param probability: The initial probability of the move leading to this node.
                            This is the 'intuition' value of the neural network.
        :param game_state: The game state of this node in the game search tree.
        """
        self.game_state = game_state
        self.config = config

        # Keep Statistics
        # P(s_prev, a)      Probability of the move leading to this node
        self.probability = probability
        # N(s)              Visit count of this node
        self.visits = 0
        # W(s) and Q(s)     Expected game outcome (total and mean)
        average_score = game_state.get_average_score()  # sane default as long as this node was not evaluated
        self.total_action_value = {player: average_score for player in game_state.get_player_list()}
        self.mean_action_value = {player: average_score for player in game_state.get_player_list()}

        # Children will be added when the node is expanded
        self.children = None

        # Keep track if we are a leave node, in that case we need to treat the game outcome different
        self.is_terminal_state = False

        # Locks for multithreaded search.
        # This is mainly to avoid 'waiting' for the neural network.
        # It also helps with exploration thanks to virtual loss.
        self.expand_lock = threading.Lock()
        self.setter_lock = threading.Lock()

        # Helpers for applying virtual losses.
        virtual_loss = game_state.get_virtual_loss()
        self._add_virtual_loss = {player: virtual_loss for player in game_state.get_player_list()}
        self._subtract_virtual_loss = {player: -virtual_loss for player in game_state.get_player_list()}

    def run_simulation_step(self, nn_client):
        """One full simulation step is selection, expansion and backup.

        This runs the three steps of the alpha zero tree search.
        It's straight forward an easy to follow with the comments.

        :return: The expected game outcome from this simulation run."""
        self._apply_virtual_loss()

        try:
            if self.is_terminal_state:
                return self._process_terminal_state()

            if self.is_expanded():
                return self._process_normal(nn_client)

            return self._process_leave(nn_client)
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise e
        finally:
            self._undo_virtual_loss()

    def _process_terminal_state(self):
        """Runs simulation step in an end game state/terminal game state.

        :return: The expected game outcome from this simulation run. As it is a terminal node this is simply the score.
        """
        self._increase_visits()
        self._update_action_value(self.game_state.calculate_scores())

        return self.game_state.calculate_scores()

    def _process_normal(self, nn_client):
        """Runs the simulation Step for a node mid tree (neither terminal, nor leave)

        :param nn_client: This client will be used to evaluate the current game state on expansion.
        :return: The expected game outcome from this simulation run.
                 This one simply returns the result backed up from lower nodes.
        """
        # Select the path to a leave node (selection step)
        move = self._select_move()
        child = self.children[move]

        # Return the result from the leave node and update own Q values (backup step)
        result = child.run_simulation_step(nn_client)
        self._increase_visits()
        self._update_action_value(result)

        return result

    def _process_leave(self, nn_client):
        """Runs simulation step in a leave node. This involves expanding the tree.

        :return: The expected game outcome from this simulation run. This is given by the neural network."""
        # We only want to expand in one thread at a time
        self.expand_lock.acquire()

        # We already expanded in another thread, simply release the lock and re-run.
        if self.is_expanded():
            self.expand_lock.release()
            return self.run_simulation_step(nn_client)

        # We actually need to expand here, lets go.
        next_states = self.game_state.get_next_game_states()
        if len(next_states) <= 0:
            # We found a terminal game state. The action value of this node is the actual score of the game state.
            self.is_terminal_state = True
            self.expand_lock.release()
            self._update_action_value(self.game_state.calculate_scores())
        else:
            # A non terminal game state will be evaluated by the neural network.
            # This gives us the expected action value for this node and predictions for move probabilities.
            evaluation = nn_client.evaluate_game_state(self.game_state)
            self._execute_expansion(evaluation.get_move_probabilities(), next_states)
            self.expand_lock.release()
            self._update_action_value(evaluation.get_expected_result())

        self._increase_visits()
        self.mean_action_value = util.deepcopy(self.total_action_value)

        # We have a expected action value for this node to back up into the tree.
        return self.mean_action_value

    def _execute_expansion(self, move_probabilities, next_states):
        if self.children:
            return

        self.children = dict()
        # Add one Node to the tree for each possible move.
        # It can be initialized with the initial move probabilities from the neural network.
        for next_state in next_states:
            move = next_state.get_last_move()
            prob = move_probabilities[move]
            self.children[move] = MCTSNode(prob, next_state, self.config)

    def _apply_virtual_loss(self):
        """Adds 'one instance' of virtual loss to this node."""
        self._increase_visits()
        self._update_action_value(self._add_virtual_loss)

    def _undo_virtual_loss(self):
        """Removes 'one instance' of virtual loss form this node."""
        self._decrease_visits()
        self._update_action_value(self._subtract_virtual_loss)

    def _select_move(self):
        """Select a move using the variant of the PUCT algorithm.
        See the alpha zero paper for a description."""

        # this equals 'sqrt(sum(visit count of children))'
        sqrt_total_child_visits = math.sqrt(self.visits)
        # constant determining exploration
        c_puct = self.config.c_puct()

        next_player = self.game_state.get_next_player()  # We need the best value for this player!
        best_move_value = -10000.0
        best_move = None
        for move, child in self.children.items():
            u = c_puct * child.probability * (sqrt_total_child_visits/(1 + child.visits))
            q = child.mean_action_value[next_player]

            move_value = u + q
            if not best_move or move_value > best_move_value:
                best_move_value = move_value
                best_move = move

        return best_move

    def _update_action_value(self, new_action_value):
        """Add the 'new_action_value' to the total and keep the average action value up to date."""
        with self.setter_lock:
            for player, value in new_action_value.items():
                self.total_action_value[player] = self.total_action_value[player] + value
                self.mean_action_value[player] = self.total_action_value[player] / max(self.visits, 1)

    def _increase_visits(self):
        with self.setter_lock:
            self.visits = self.visits + 1

    def _decrease_visits(self):
        with self.setter_lock:
            self.visits = self.visits - 1

    def is_expanded(self):
        return not not self.children

    def move_probabilities(self, temperature):
        """Returns each possible move and its probability. Temperature controls how much extreme values are damped.

        See the Alpha Go Zero paper for more details. Usually temperature is simply 1 to return the
        number of visits for each subtree."""
        exponent = 1.0 / temperature

        visit_sum = 0
        exponentiated_visit_counts = dict()
        for move, child in self.children.items():
            exponentiated_visit_count = child.visits ** exponent
            visit_sum = visit_sum + exponentiated_visit_count

            exponentiated_visit_counts[move] = exponentiated_visit_count

        if visit_sum > 0:
            return {move: count / visit_sum for move, count in exponentiated_visit_counts.items()}
        else:
            raise Exception('Must run simulations before getting move probabilities.')
